# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from quantrocket.cli.utils.parse import dict_str, list_or_int_or_float_or_str

def add_subparser(subparsers):
    _parser = subparsers.add_parser("moonshot", description="QuantRocket Moonshot CLI", help="Backtest and trade Moonshot strategies")
    _subparsers = _parser.add_subparsers(title="subcommands", dest="subcommand")
    _subparsers.required = True

    examples = """
Backtest one or more strategies.

By default returns a CSV of backtest results but can also return a PDF tear sheet of
performance charts.

If testing multiple strategies, each column in the CSV represents a strategy.
If testing a single strategy with the `--details` option, each column in the CSV
represents a security in the strategy universe.

Examples:

Backtest several HML (High Minus Low) strategies from 2005-2015 and return a
CSV of results:

    quantrocket moonshot backtest hml-us hml-eur hml-asia -s 2005-01-01 -e 2015-12-31 -o hml_results.csv

Backtest a single strategy called demo using all available history and return a
PDF tear sheet:

    quantrocket moonshot backtest demo --pdf -o tearsheet.pdf

Run a backtest in 1-year segments to reduce memory usage:

    quantrocket moonshot backtest big-strategy -s 2000-01-01 -e 2018-01-01 --segment A -o results.csv
    """
    parser = _subparsers.add_parser(
        "backtest",
        help="backtest one or more strategies",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "strategies",
        nargs="+",
        metavar="CODE",
        help="one or more strategy codes")
    backtest_options = parser.add_argument_group("backtest options")
    backtest_options.add_argument(
        "-s", "--start-date",
        metavar="YYYY-MM-DD",
        help="the backtest start date (default is to use all available history)")
    backtest_options.add_argument(
        "-e", "--end-date",
        metavar="YYYY-MM-DD",
        help="the backtest end date (default is to use all available history)")
    backtest_options.add_argument(
        "-g", "--segment",
        metavar="FREQ",
        help="backtest in date segments of this size, to reduce memory usage "
        "(use Pandas frequency string, e.g. 'A' for annual segments or 'Q' "
        "for quarterly segments)")
    backtest_options.add_argument(
        "-l", "--allocations",
        type=dict_str,
        metavar="CODE:FLOAT",
        nargs="*",
        help="the allocation for each strategy, passed as 'code:allocation' (default "
        "allocation is 1.0 / number of strategies)")
    backtest_options.add_argument(
        "-n", "--nlv",
        nargs="*",
        type=dict_str,
        metavar="CURRENCY:NLV",
        help="the NLV (net liquidation value, i.e. account balance) to assume for "
        "the backtest, expressed in each currency represented in the backtest (pass "
        "as 'currency:nlv')")
    backtest_options.add_argument(
        "-p", "--params",
        nargs="*",
        type=dict_str,
        metavar="PARAM:VALUE",
        help="one or more strategy params to set on the fly before backtesting "
        "(pass as 'param:value')")
    backtest_options.add_argument(
        "--no-cache",
        action="store_true",
        help="don't use cached files even if available. Using cached files speeds "
        "up backtests but may be undesirable if underlying data has changed. "
        "See http://qrok.it/h/mcache to learn more about caching in Moonshot.")
    outputs = parser.add_argument_group("output options")
    outputs.add_argument(
        "-d", "--details",
        action="store_true",
        help="return detailed results for all securities instead of aggregating to "
        "strategy level (only supported for single-strategy backtests)")
    outputs.add_argument(
        "--pdf",
        action="store_const",
        const="pdf",
        dest="output",
        help="return a PDF performance tear sheet (default is to return a CSV "
        "of performance results)")
    outputs.add_argument(
        "-o", "--outfile",
        metavar="FILEPATH",
        dest="filepath_or_buffer",
        help="the location to write the results file (omit to write to stdout)")
    parser.set_defaults(func="quantrocket.moonshot._cli_backtest")

    examples="""
Run a parameter scan for one or more strategies.

By default returns a CSV of scan results but can also return a PDF tear sheet.

Examples:

Run a parameter scan for several different moving averages on a strategy
called trend-friend and return a PDF:

    quantrocket moonshot paramscan trend-friend -p MAVG_WINDOW -v 20 50 100 --pdf -o tearsheet.pdf

Run a 2-D parameter scan for multiple strategies and return a PDF:

    quantrocket moonshot paramscan strat1 strat2 strat3 -p MIN_STD -v 1 1.5 2 --param2 STD_WINDOW --vals2 20 50 100 200 --pdf -o tearsheet.pdf

Run a parameter scan in 1-year segments to reduce memory usage:

    quantrocket moonshot paramscan big-strategy -s 2000-01-01 -e 2018-01-01 --segment A -p MAVG_WINDOW -v 20 50 100 --pdf -o tearsheet.pdf
    """
    parser = _subparsers.add_parser(
        "paramscan",
        help="run a parameter scan for one or more strategies",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "strategies",
        nargs="+",
        metavar="CODE",
        help="one or more strategy codes")
    backtest_options = parser.add_argument_group("backtest options")
    backtest_options.add_argument(
        "-s", "--start-date",
        metavar="YYYY-MM-DD",
        help="the backtest start date (default is to use all available history)")
    backtest_options.add_argument(
        "-e", "--end-date",
        metavar="YYYY-MM-DD",
        help="the backtest end date (default is to use all available history)")
    backtest_options.add_argument(
        "-g", "--segment",
        metavar="FREQ",
        help="backtest in date segments of this size, to reduce memory usage "
        "(use Pandas frequency string, e.g. 'A' for annual segments or 'Q' "
        "for quarterly segments)")
    backtest_options.add_argument(
        "-p", "--param1",
        metavar="PARAM",
        type=str,
        required=True,
        help="the name of the parameter to test (a class attribute on the strategy)")
    backtest_options.add_argument(
        "-v", "--vals1",
        type=list_or_int_or_float_or_str,
        metavar="VALUE",
        nargs="+",
        required=True,
        help="parameter values to test (values can be integers, floats, strings, 'True', "
        "'False', 'None', or 'default' (to test current param value); for lists/tuples, "
        "use comma-separated values)")
    backtest_options.add_argument(
        "--param2",
        metavar="PARAM",
        type=str,
        help="name of a second parameter to test (for 2-D parameter scans)")
    backtest_options.add_argument(
        "--vals2",
        type=list_or_int_or_float_or_str,
        metavar="VALUE",
        nargs="*",
        help="values to test for parameter 2 (values can be integers, floats, strings, "
        "'True', 'False', 'None', or 'default' (to test current param value); for "
        "lists/tuples, use comma-separated values)")
    backtest_options.add_argument(
        "-l", "--allocations",
        type=dict_str,
        metavar="CODE:FLOAT",
        nargs="*",
        help="the allocation for each strategy, passed as 'code:allocation' (default "
        "allocation is 1.0 / number of strategies)")
    backtest_options.add_argument(
        "-n", "--nlv",
        nargs="*",
        type=dict_str,
        metavar="CURRENCY:NLV",
        help="the NLV (net liquidation value, i.e. account balance) to assume for "
        "the backtests, expressed in each currency represented in the backtest (pass "
        "as 'currency:nlv')")
    backtest_options.add_argument(
        "--params",
        nargs="*",
        type=dict_str,
        metavar="PARAM:VALUE",
        help="one or more strategy params to set on the fly before backtesting "
        "(pass as 'param:value')")
    backtest_options.add_argument(
        "--no-cache",
        action="store_true",
        help="don't use cached files even if available. Using cached files speeds "
        "up backtests but may be undesirable if underlying data has changed. "
        "See http://qrok.it/h/mcache to learn more about caching in Moonshot.")
    outputs = parser.add_argument_group("output options")
    outputs.add_argument(
        "--pdf",
        action="store_const",
        const="pdf",
        dest="output",
        help="return a PDF tear sheet of results (default is to return a CSV)")
    outputs.add_argument(
        "-o", "--outfile",
        metavar="FILEPATH",
        dest="filepath_or_buffer",
        help="the location to write the results file (omit to write to stdout)")
    parser.set_defaults(func="quantrocket.moonshot._cli_scan_parameters")

    examples = """
Run a walk-forward optimization of a machine learning strategy.

The date range will be split into segments of `--train` size. For each
segment, the model will be trained with the data, then the trained model will
be backtested on the following segment.

By default, uses scikit-learn's StandardScaler+SGDRegressor. Also supports other
scikit-learn models/pipelines and Keras models. To customize model, instantiate
the model locally, serialize it to disk, and pass the filename of the serialized
model as `--model`.

Supports expanding walk-forward optimizations (the default), which use an anchored start date
for model training, or rolling walk-forward optimizations (by specifying `--rolling-train`),
which use a rolling or non-anchored start date for model training.

Returns a backtest results CSV and a dump of the machine learning model
as of the end of the analysis.

Examples:

Run a walk-forward optimization using the default model and retrain the model annually,
writing the backtest results and trained model to demo_ml_results.csv and
demo_ml_trained_model.joblib, respectively:

    quantrocket moonshot ml-walkforward demo-ml -s 2007-01-01 -e 2018-12-31 --train A -o demo_ml*

Run a walk-forward optimization using a custom model (serialized with joblib), retrain the
model annually, don't perform backtesting until after 5 years of initial training,
and further split the training and backtesting into quarterly segments to reduce
memory usage:

    quantrocket moonshot ml-walkforward demo-ml -s 2007-01-01 -e 2018-12-31 --model my_model.joblib --train A --min-train 5Y --segment Q -o demo_ml*
    """
    parser = _subparsers.add_parser(
        "ml-walkforward",
        help="run a walk-forward optimization of a machine learning strategy",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "strategy",
        metavar="CODE",
        help="the strategy code")
    walkforward_options = parser.add_argument_group("walk-forward analysis options")
    walkforward_options.add_argument(
        "-s", "--start-date",
        metavar="YYYY-MM-DD",
        required=True,
        help="the analysis start date (note that model training will start on this date "
        "but backtesting will not start until after the initial training period)")
    walkforward_options.add_argument(
        "-e", "--end-date",
        metavar="YYYY-MM-DD",
        required=True,
        help="the analysis end date")
    walkforward_options.add_argument(
        "-t", "--train",
        metavar="FREQ",
        required=True,
        help="train model this frequently (use Pandas frequency string, e.g. 'A' "
        "for annual training or 'Q' for quarterly training)")
    walkforward_options.add_argument(
        "-m", "--min-train",
        metavar="FREQ",
        help="don't backtest until at least this much model training has occurred; "
        "defaults to the length of `--train` if not specified (use Pandas frequency "
        "string, e.g. '5Y' for 5 years of initial training)")
    walkforward_options.add_argument(
        "-r", "--rolling-train",
        metavar="FREQ",
        help="train model with a rolling window of this length; if omitted, train "
        "model with an expanding window (use Pandas frequency string, e.g. '3Y' for "
        "a 3-year rolling training window)")
    walkforward_options.add_argument(
        "-f", "--model",
        dest="model_filepath",
        help="filepath of serialized model to use, filename must end in '.joblib' or "
        "'.pkl' (if omitted, default model is scikit-learn's StandardScaler+SGDRegressor)")
    walkforward_options.add_argument(
        "--force-nonincremental",
        action="store_true",
        help="force the model to be trained non-incrementally (i.e. load entire training "
        "data set into memory) even if it supports incremental learning. Required "
        "in order to perform a rolling (as opposed to expanding) walk-forward optimization "
        "with a model that supports incremental learning.")
    backtest_options = parser.add_argument_group("backtest options")
    backtest_options.add_argument(
        "-g", "--segment",
        metavar="FREQ",
        help="train and backtest in date segments of this size, to reduce memory "
        "usage; must be smaller than `--train`/`--min-train` or will have no effect "
        "(use Pandas frequency string, e.g. 'A' for annual segments or 'Q' for "
        "quarterly segments)")
    backtest_options.add_argument(
        "-l", "--allocation",
        type=float,
        metavar="FLOAT",
        help="the allocation for the strategy (default 1.0)")
    backtest_options.add_argument(
        "-n", "--nlv",
        nargs="*",
        type=dict_str,
        metavar="CURRENCY:NLV",
        help="the NLV (net liquidation value, i.e. account balance) to assume for "
        "the backtest, expressed in each currency represented in the backtest (pass "
        "as 'currency:nlv')")
    backtest_options.add_argument(
        "-p", "--params",
        nargs="*",
        type=dict_str,
        metavar="PARAM:VALUE",
        help="one or more strategy params to set on the fly before backtesting "
        "(pass as 'param:value')")
    backtest_options.add_argument(
        "--no-cache",
        action="store_true",
        help="don't use cached files even if available. Using cached files speeds "
        "up backtests but may be undesirable if underlying data has changed. "
        "See http://qrok.it/h/mcache to learn more about caching in Moonshot.")
    outputs = parser.add_argument_group("output options")
    outputs.add_argument(
        "-d", "--details",
        action="store_true",
        help="return detailed results for all securities instead of aggregating")
    outputs.add_argument(
        "--progress",
        action="store_true",
        help="log status and Sharpe ratios of each walk-forward segment during "
        "analysis (default False)")
    outputs.add_argument(
        "-o", "--outfile",
        metavar="FILEPATH",
        dest="filepath_or_buffer",
        help="the location to write the ZIP file to; or, if path ends with '*', the "
        "pattern to use for extracting the zipped files. For example, if the path is "
        "my_ml*, files will extracted to my_ml_results.csv and my_ml_trained_model.joblib.")
    parser.set_defaults(func="quantrocket.moonshot._cli_ml_walkforward")

    examples = """
Run one or more strategies and generate orders.

Allocations are read from configuration (quantrocket.moonshot.allocations.yml).

Examples:

Generate orders for a single strategy called umd-nyse:

    quantrocket moonshot trade umd-nyse -o orders.csv

Generate orders and automatically place them (if any) through the blotter:

    quantrocket moonshot trade umd-nyse | quantrocket blotter order -f -

Generate orders for multiple strategies for a particular account:

    quantrocket moonshot trade umd-japan hml-japan --accounts DU12345 -o orders.csv

Generate orders as if it were an earlier date (for prupose of review):

    quantrocket moonshot trade umd-nyse -o orders.csv --review-date 2018-05-11
    """
    parser = _subparsers.add_parser(
        "trade",
        help="run one or more strategies and generate orders.",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "strategies",
        nargs="+",
        metavar="CODE",
        help="one or more strategy codes")
    parser.add_argument(
        "-a", "--accounts",
        metavar="ACCOUNT",
        nargs="*",
        help="limit to these accounts")
    parser.add_argument(
         "-r", "--review-date",
         metavar="YYYY-MM-DD",
         help="generate orders as if it were this date, rather than using today's date")
    parser.add_argument(
        "-j", "--json",
        action="store_const",
        const="json",
        dest="output",
        help="format orders as JSON (default is CSV)")
    parser.add_argument(
        "-o", "--outfile",
        metavar="FILEPATH",
        dest="filepath_or_buffer",
        help="the location to write the orders file (omit to write to stdout)")
    parser.set_defaults(func="quantrocket.moonshot._cli_trade")

    #parser = _subparsers.add_parser("walkforward", help="run walkforward analysis for one or more strategies")
    #parser.add_argument("start_date", metavar="YYYY-MM-DD", help="start date")
    #parser.add_argument("end_date", metavar="YYYY-MM-DD", help="end date")
    #parser.add_argument("-s", "--strategies", nargs="+", metavar="CODE", help="one or more strategies to backtest")
    #parser.add_argument("-l", "--allocations", type=float, metavar="FLOAT", nargs="*", help="the allocations for each strategy, if different from the registered allocation (must be the same length as -s/--strategies if provided)")
    #parser.add_argument("-i", "--interval", metavar="OFFSET", required=True, help="walkforward intervals as a Pandas offset string (e.g. MS, QS, 6MS, AS, see http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases)")
    #parser.add_argument("-p", "--param1", metavar="PARAM", type=str, required=True, help="name of the parameter to test")
    #parser.add_argument("-v", "--vals1", metavar="VALUE", nargs="+", help="parameter values to test")
    #parser.add_argument("-f", "--func1", metavar="PATH", help="dot-separated path of a function with which to transform the test values")
    #parser.add_argument("--rolling", action="store_true", help="use a rolling window to calculate performance (default is to use an expanding window)")
    #parser.add_argument("-r", "--rankby", choices=["cagr", "sharpe"], help="rank each period's performance by 'sharpe', 'cagr', or a 'blend' of both (default blend)")
    #parser.add_argument("-a", "--account", help="use the latest NLV of this account for modeling commissions, liquidity constraints, etc.")
    #parser.add_argument("--params", nargs="*", metavar="PARAM:VALUE", help="strategy params to set on the fly (distinct from the params to be scanned)")
    #parser.add_argument("-w", "--raw", action="store_true", help="return raw performance data instead of a performance tearsheet")
    #parser.set_defaults(func="quantrocket.moonshot.walkforward")

    #parser = _subparsers.add_parser("params", help="view params for one or more strategies")
    #parser.add_argument("codes", nargs="+", metavar="CODE", help="the strategies to include")
    #parser.add_argument("-d", "--diff", action="store_true", help="exclude params that are the same for all strategies")
    #parser.add_argument("-p", "--params", metavar="PARAM", nargs="*", help="limit to these params")
    #parser.add_argument("-g", "--group-by-class", choices=["parent", "child"], help="""
    #group by the topmost ("parent") or bottommost ("child") class in the strategy class hierarchy
    #where the param is defined (default no class grouping). (Use "parent" to group by category
    #and "child" to show where param would need to be edited.)""")
    #parser.add_argument("-s", "--groupsort", action="store_true", help="sort by origin class (requires -g/--group-by-class), otherwise sort by param name")
    #parser.set_defaults(func="quantrocket.moonshot.get_params")

    #parser = _subparsers.add_parser("shortfall", help="compare live and simulated results")
    #parser.add_argument("start_date", metavar="YYYY-MM-DD", help="start date")
    #parser.add_argument("end_date", nargs="?", metavar="YYYY-MM-DD", help="end date (optional)")
    #parser.add_argument("-s", "--strategies", nargs="+", metavar="CODE", help="one or more strategies to show shortfall for")
    #parser.add_argument("-l", "--allocations", type=float, metavar="FLOAT", nargs="*", help="the allocations for each strategy, if different from the registered allocation (must be the same length as -s/--strategies if provided)")
    #parser.add_argument("-a", "--account", help="the account to compare shortfall for (if not provided, the default account registered with the account service will be used)")
    #parser.add_argument("-p", "--params", nargs="*", metavar="PARAM:VALUE", help="strategy params to set on the fly")
    #parser.add_argument("-w", "--raw", action="store_true", help="return raw performance data instead of a performance tearsheet")
    #parser.set_defaults(func="quantrocket.moonshot.get_implementation_shortfall")



# Copyright 2020 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import six
import json
from quantrocket.houston import houston
from quantrocket.cli.utils.output import json_to_cli
from quantrocket.cli.utils.stream import to_bytes
from quantrocket.cli.utils.files import write_response_to_filepath_or_buffer
from quantrocket.exceptions import ParameterError

def list_ibkr_exchanges(regions=None, sec_types=None):
    """
    List exchanges by security type and country as found on the IBKR website.

    Parameters
    ----------
    regions : list of str, optional
        limit to these regions. Possible choices: north_america, europe, asia, global

    sec_types : list of str, optional
        limit to these securitiy types. Possible choices: STK, ETF, FUT, CASH, IND

    Returns
    -------
    dict
    """
    params = {}
    if sec_types:
        params["sec_types"] = sec_types
    if regions:
        params["regions"] = regions

    response = houston.get("/master/exchanges/ibkr", params=params, timeout=180)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_ibkr_exchanges(*args, **kwargs):
    return json_to_cli(list_ibkr_exchanges, *args, **kwargs)

def collect_alpaca_listings():
    """
    Collect securities listings from Alpaca and store in securities master
    database.

    Returns
    -------
    dict
        status message
    """
    response = houston.post("/master/securities/alpaca")
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_alpaca_listings(*args, **kwargs):
    return json_to_cli(collect_alpaca_listings, *args, **kwargs)

def collect_edi_listings(exchanges=None):
    """
    Collect securities listings from EDI and store in securities master
    database.

    Parameters
    ----------
    exchanges : list or str, required
        collect listings for these exchanges (identified by MICs)

    Returns
    -------
    dict
        status message

    Examples
    --------
    Collect sample listings:

    >>> collect_edi_listings(exchanges="FREE")

    Collect listings for all permitted exchanges:

    >>> collect_edi_listings()

    Collect all Chinese stock listings:

    >>> collect_edi_listings(exchanges=["XSHG", "XSHE"])
    """
    params = {}
    if exchanges:
        params["exchanges"] = exchanges

    response = houston.post("/master/securities/edi", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_edi_listings(*args, **kwargs):
    return json_to_cli(collect_edi_listings, *args, **kwargs)

def collect_figi_listings():
    """
    Collect securities listings from Bloomberg OpenFIGI and store
    in securities master database.

    OpenFIGI provides several useful security attributes including
    market sector, a detailed security type, and share class-level
    FIGI identifier.

    The collected data fields show up in the master file with the
    prefix "figi_*".

    This function does not directly query the OpenFIGI API but rather
    downloads a dump of all FIGIs which QuantRocket has previously
    mapped to securities from other vendors.

    Returns
    -------
    dict
        status message

    Examples
    --------
    Collect all available FIGI listings:

    >>> collect_figi_listings()
    """
    response = houston.post("/master/securities/figi")
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_figi_listings(*args, **kwargs):
    return json_to_cli(collect_figi_listings, *args, **kwargs)

def collect_ibkr_listings(exchanges=None, sec_types=None, currencies=None,
                          symbols=None, universes=None, sids=None):
    """
    Collect securities listings from Interactive Brokers and store in
    securities master database.

    Specify an exchange (optionally filtering by security type, currency,
    and/or symbol) to collect listings from the IBKR website and collect
    associated contract details from the IBKR API. Or, specify universes or
    sids to collect details from the IBKR API, bypassing the website.

    Parameters
    ----------
    exchanges : list or str
        one or more IBKR exchange codes to collect listings for (required
        unless providing universes or sids). For sample data use exchange
        code 'FREE'

    sec_types : list of str, optional
        limit to these security types. Possible choices: STK, ETF, FUT, CASH, IND

    currencies : list of str, optional
        limit to these currencies

    symbols : list of str, optional
        limit to these symbols

    universes : list of str, optional
        limit to these universes

    sids : list of str, optional
        limit to these sids

    Returns
    -------
    dict
        status message

    Examples
    --------
    Collect free sample listings:

    >>> collect_ibkr_listings(exchanges="FREE")

    Collect all Toronto Stock Exchange stock listings:

    >>> collect_ibkr_listings(exchanges="TSE", sec_types="STK")

    Collect all NYSE ARCA ETF listings:

    >>> collect_ibkr_listings(exchanges="ARCA", sec_types="ETF")

    Collect specific symbols from Nasdaq:

    >>> collect_ibkr_listings(exchanges="NASDAQ", symbols=["AAPL", "GOOG", "NFLX"])

    Re-collect contract details for an existing universe called "japan-fin":

    >>> collect_ibkr_listings(universes="japan-fin")
    """
    params = {}
    if exchanges:
        params["exchanges"] = exchanges
    if sec_types:
        params["sec_types"] = sec_types
    if currencies:
        params["currencies"] = currencies
    if symbols:
        params["symbols"] = symbols
    if universes:
        params["universes"] = universes
    if sids:
        params["sids"] = sids

    response = houston.post("/master/securities/ibkr", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_ibkr_listings(*args, **kwargs):
    return json_to_cli(collect_ibkr_listings, *args, **kwargs)

def collect_sharadar_listings(countries="US"):
    """
    Collect securities listings from Sharadar and store in securities master
    database.

    Parameters
    ----------
    countries : list of str, required
        countries to collect listings for. Possible choices: US, FREE

    Returns
    -------
    dict
        status message
    """
    params = {}
    if countries:
        params["countries"] = countries

    response = houston.post("/master/securities/sharadar", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_sharadar_listings(*args, **kwargs):
    return json_to_cli(collect_sharadar_listings, *args, **kwargs)

def collect_usstock_listings():
    """
    Collect US stock listings from QuantRocket and store in securities
    master database.

    Returns
    -------
    dict
        status message
    """
    response = houston.post("/master/securities/usstock")
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_usstock_listings(*args, **kwargs):
    return json_to_cli(collect_usstock_listings, *args, **kwargs)

def collect_ibkr_option_chains(universes=None, sids=None, infilepath_or_buffer=None):
    """
    Collect IBKR option chains for underlying securities.

    Note: option chains often consist of hundreds, sometimes thousands of
    options per underlying security. Be aware that requesting option chains
    for large universes of underlying securities, such as all stocks on the
    NYSE, can take numerous hours to complete.

    Parameters
    ----------
    universes : list of str, optional
        collect options for these universes of underlying securities

    sids : list of str, optional
        collect options for these underlying sids

    infilepath_or_buffer : str or file-like object, optional
        collect options for the sids in this file (specify '-' to read file
        from stdin)

    Returns
    -------
    dict
        status message
    """
    params = {}
    if universes:
        params["universes"] = universes
    if sids:
        params["sids"] = sids

    if infilepath_or_buffer == "-":
        response = houston.post("/master/options/ibkr", params=params, data=to_bytes(sys.stdin))

    elif infilepath_or_buffer and hasattr(infilepath_or_buffer, "read"):
        if infilepath_or_buffer.seekable():
            infilepath_or_buffer.seek(0)
        response = houston.post("/master/options/ibkr", params=params, data=to_bytes(infilepath_or_buffer))

    elif infilepath_or_buffer:
        with open(infilepath_or_buffer, "rb") as f:
            response = houston.post("/master/options/ibkr", params=params, data=f)
    else:
        response = houston.post("/master/options/ibkr", params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_ibkr_option_chains(*args, **kwargs):
    return json_to_cli(collect_ibkr_option_chains, *args, **kwargs)

def diff_ibkr_securities(universes=None, sids=None, infilepath_or_buffer=None,
                         fields=None, delist_missing=False, delist_exchanges=None,
                         wait=False):
    """
    Flag security details that have changed in IBKR's system since the time they
    were last collected into the securities master database.

    Diff can be run synchronously or asynchronously (asynchronous is the default
    and is recommended if diffing more than a handful of securities).

    Parameters
    ----------
    universes : list of str, optional
        limit to these universes

    sids : list of str, optional
        limit to these sids

    infilepath_or_buffer : str or file-like object, optional
        limit to the sids in this file (specify '-' to read file from stdin)

    fields : list of str, optional
        only diff these fields (field name should start with "ibkr")

    delist_missing : bool
        auto-delist securities that are no longer available from IBKR

    delist_exchanges : list of str, optional
        auto-delist securities that are associated with these exchanges

    wait : bool
        run the diff synchronously and return the diff (otherwise run
        asynchronously and log the results, if any, to flightlog)

    Returns
    -------
    dict
        dict of sids and fields that have changed (if wait), or status message

    """
    params = {}
    if universes:
        params["universes"] = universes
    if sids:
        params["sids"] = sids
    if fields:
        params["fields"] = fields
    if delist_missing:
        params["delist_missing"] = delist_missing
    if delist_exchanges:
        params["delist_exchanges"] = delist_exchanges
    if wait:
        params["wait"] = wait

    # if run synchronously use a high timeout
    timeout = 60*60*10 if wait else None
    if infilepath_or_buffer == "-":
        response = houston.get("/master/diff/ibkr", params=params, data=to_bytes(sys.stdin), timeout=timeout)

    elif infilepath_or_buffer and hasattr(infilepath_or_buffer, "read"):
        if infilepath_or_buffer.seekable():
            infilepath_or_buffer.seek(0)
        response = houston.get("/master/diff/ibkr", params=params, data=to_bytes(infilepath_or_buffer), timeout=timeout)

    elif infilepath_or_buffer:
        with open(infilepath_or_buffer, "rb") as f:
            response = houston.get("/master/diff/ibkr", params=params, data=f, timeout=timeout)
    else:
        response = houston.get("/master/diff/ibkr", params=params, timeout=timeout)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_diff_ibkr_securities(*args, **kwargs):
    return json_to_cli(diff_ibkr_securities, *args, **kwargs)

def download_master_file(filepath_or_buffer=None, output="csv", exchanges=None, sec_types=None,
                         currencies=None, universes=None, symbols=None, sids=None,
                         exclude_universes=None, exclude_sids=None,
                         exclude_delisted=False, exclude_expired=False,
                         frontmonth=False, vendors=None, fields=None):
    """
    Query security details from the securities master database and download to file.

    Parameters
    ----------
    filepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    output : str
        output format (csv or json, default is csv)

    exchanges : list of str, optional
        limit to these exchanges. You can specify exchanges using the MIC or the
        vendor's exchange code.

    sec_types : list of str, optional
        limit to these security types. Possible choices: STK, ETF, FUT, CASH, IND, OPT, FOP, BAG

    currencies : list of str, optional
        limit to these currencies

    universes : list of str, optional
        limit to these universes

    symbols : list of str, optional
        limit to these symbols

    sids : list of str, optional
        limit to these sids

    exclude_universes : list of str, optional
        exclude these universes

    exclude_sids : list of str, optional
        exclude these sids

    exclude_delisted : bool
        exclude delisted securities (default is to include them)

    exclude_expired : bool
        exclude expired contracts (default is to include them)

    frontmonth : bool
        exclude backmonth and expired futures contracts (default False)

    vendors : list of str, optional
        limit to these vendors. Possible choices: alpaca, edi, ibkr,
        sharadar, usstock

    fields : list of str, optional
        Return specific fields. By default a core set of fields is
        returned, but additional vendor-specific fields are also available.
        To return non-core fields, you can reference them by name, or pass "*"
        to return all available fields. To return all fields for a specific
        vendor, pass the vendor prefix followed by "*", for example "edi*"
        for all EDI fields. Pass "?*" (or any invalid vendor prefix plus "*")
        to see available vendor prefixes. Pass "?" or any invalid fieldname
        to see all available fields.

    Returns
    -------
    None

    Examples
    --------
    Download NYSE and NASDAQ securities to file, using MICs to specify
    the exchanges:

    >>> download_master_file("securities.csv", exchanges=["XNYS","XNAS"])

    Download NYSE and NASDAQ securities to file, using IBKR exchange codes
    to specify the exchanges, and include all IBKR fields:

    >>> download_master_file("securities.csv", exchanges=["NYSE","NASDAQ"], fields="ibkr*")

    Download securities for a particular universe to in-memory file, including
    all possible fields, and load the CSV into pandas.

    >>> f = io.StringIO()
    >>> download_master_file(f, fields="*", universes="my-universe")
    >>> securities = pd.read_csv(f)
    """
    params = {}
    if exchanges:
        params["exchanges"] = exchanges
    if sec_types:
        params["sec_types"] = sec_types
    if currencies:
        params["currencies"] = currencies
    if universes:
        params["universes"] = universes
    if symbols:
        params["symbols"] = symbols
    if sids:
        params["sids"] = sids
    if exclude_universes:
        params["exclude_universes"] = exclude_universes
    if exclude_sids:
        params["exclude_sids"] = exclude_sids
    if exclude_delisted:
        params["exclude_delisted"] = exclude_delisted
    if exclude_expired:
        params["exclude_expired"] = exclude_expired
    if frontmonth:
        params["frontmonth"] = frontmonth
    if vendors:
        params["vendors"] = vendors
    if fields:
        params["fields"] = fields

    output = output or "csv"

    url = "/master/securities.{0}".format(output)

    if output not in ("csv", "json"):
        raise ValueError("Invalid ouput: {0}".format(output))

    response = houston.get(url, params=params)

    houston.raise_for_status_with_json(response)

    filepath_or_buffer = filepath_or_buffer or sys.stdout

    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_download_master_file(*args, **kwargs):
    return json_to_cli(download_master_file, *args, **kwargs)

def get_securities_reindexed_like(reindex_like, fields=None):
    """
    Return a multiindex DataFrame of securities master data, reindexed to
    match the index and columns (sids) of `reindex_like`.

    Parameters
    ----------
    reindex_like : DataFrame, required
        a DataFrame (usually of prices) with dates for the index and sids
        for the columns, to which the shape of the resulting DataFrame will
        be conformed

    fields : list of str
        a list of fields to include in the resulting DataFrame. By default a
        core set of fields is returned, but additional vendor-specific fields
        are also available. To return non-core fields, you can reference them
        by name, or pass "*" to return all available fields. To return all
        fields for a specific vendor, pass the vendor prefix followed by "*",
        for example "edi*" for all EDI fields. Pass "?*" (or any invalid
        vendor prefix plus "*") to see available vendor prefixes. Pass "?" or
        any invalid fieldname to see all available fields.

    Returns
    -------
    DataFrame
        a multiindex (Field, Date) DataFrame of securities master data, shaped
        like the input DataFrame

    Examples
    --------
    Get exchanges (MICs) using a DataFrame of prices:

    >>> closes = prices.loc["Close"]
    >>> securities = get_securities_reindexed_like(
            closes, fields=["Exchange"])
    >>> exchanges = securities.loc["Exchange"]
    >>> nyse_closes = closes.where(exchanges == "XNYS")
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function")

    sids = list(reindex_like.columns)

    f = six.StringIO()
    download_master_file(f, sids=sids, fields=fields)
    securities = pd.read_csv(f, index_col="Sid")

    if "Sid" in fields:
        securities["Sid"] = securities.index

    all_master_fields = {}

    for col in securities.columns:
        this_col = securities[col]
        if col.endswith("Delisted") or col.endswith("Etf"):
            this_col = this_col.fillna(0).astype(bool)
        all_master_fields[col] = reindex_like.apply(lambda x: this_col, axis=1)

    names = list(reindex_like.index.names)
    names.insert(0, "Field")

    securities = pd.concat(all_master_fields, names=names)
    securities = securities.reindex(columns=reindex_like.columns)
    return securities

def get_contract_nums_reindexed_like(reindex_like, limit=5):
    """
    From a DataFrame of futures (with dates as the index and sids as columns),
    return a DataFrame of integers representing each sid's sequence in the
    futures chain as of each date, where 1 is the front contract, 2 is the second
    nearest contract, etc.

    Sequences are based on the RolloverDate field in the securities master
    file, which is based on configurable rollover rules.

    Parameters
    ----------
    reindex_like : DataFrame, required
        a DataFrame (usually of prices) with dates for the index and sids
        for the columns, to which the shape of the resulting DataFrame will
        be conformed

    limit : int
        how many contracts ahead to sequence. For example, assuming quarterly
        contracts, a limit of 5 will sequence 5 quarters out. Default 5.

    Returns
    -------
    DataFrame
        a DataFrame of futures chain sequence numbers, shaped like the input
        DataFrame

    Examples
    --------
    Get a Boolean mask of front-month contracts:

    >>> closes = prices.loc["Close"]
    >>> contract_nums = get_contract_nums_reindexed_like(closes)
    >>> are_front_months = contract_nums == 1
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function")

    index_levels = reindex_like.index.names

    if "Date" not in index_levels:
        raise ParameterError(
            "reindex_like must have index called 'Date', but has {0}".format(
                ",".join([str(name) for name in index_levels])))

    reindex_like_dt_index = reindex_like.index.get_level_values("Date")

    if not hasattr(reindex_like_dt_index, "date"):
        raise ParameterError("reindex_like must have a DatetimeIndex")

    f = six.StringIO()
    download_master_file(f, sids=list(reindex_like.columns),
                         fields=["RolloverDate","ibkr_UnderConId","SecType"])
    rollover_dates = pd.read_csv(f, parse_dates=["RolloverDate"])
    rollover_dates = rollover_dates[rollover_dates.SecType=="FUT"].drop("SecType", axis=1)

    if rollover_dates.empty:
        raise ParameterError("input DataFrame does not appear to contain any futures contracts")

    if reindex_like_dt_index.tz:
        rollover_dates.loc[:, "RolloverDate"] = rollover_dates.RolloverDate.dt.tz_localize(reindex_like_dt_index.tz.zone)

    min_date = reindex_like_dt_index.min()
    max_date = max([rollover_dates.RolloverDate.max(),
                    reindex_like_dt_index.max()])

    # Stack sids by underlying (1 column per underlying)
    rollover_dates = rollover_dates.set_index(["RolloverDate","ibkr_UnderConId"]).Sid.unstack()

    contract_nums = None

    for i in range(limit):

        # backshift conids
        _rollover_dates = rollover_dates.apply(lambda col: col.dropna().shift(-i))

        # Reindex to daily frequency
        _rollover_dates = _rollover_dates.reindex(
            index=pd.date_range(start=min_date, end=max_date))

        # RolloverDate is when we roll out of the contract, hence we backfill
        _rollover_dates = _rollover_dates.fillna(method="bfill")

        # Stack to Series of Date, nth sid
        _rollover_dates = _rollover_dates.stack()
        _rollover_dates.index = _rollover_dates.index.droplevel("ibkr_UnderConId")
        _rollover_dates.index.name = "Date"

        # Pivot Series to DataFrame
        _rollover_dates = _rollover_dates.reset_index(name="Sid")
        _rollover_dates["ContractNum"] = i + 1
        _rollover_dates = _rollover_dates.set_index(["Date","Sid"])
        _contract_nums = _rollover_dates.ContractNum.unstack()

        # If MultiIndex input, broadcast across Time level
        if len(index_levels) > 1:
            _contract_nums = _contract_nums.reindex(index=reindex_like.index,
                                                level="Date")
            _contract_nums = _contract_nums.reindex(columns=reindex_like.columns)
        else:
            _contract_nums = _contract_nums.reindex(index=reindex_like_dt_index,
                                                    columns=reindex_like.columns)

        if contract_nums is None:
            contract_nums = _contract_nums
        else:
            contract_nums = contract_nums.fillna(_contract_nums)

    return contract_nums

def create_universe(code, infilepath_or_buffer=None, from_universes=None,
                    exclude_delisted=False, append=False, replace=False):
    """
    Create a universe of securities.

    Parameters
    ----------
    code : str, required
        the code to assign to the universe (lowercase alphanumerics and hyphens only)

    infilepath_or_buffer : str or file-like object, optional
        create the universe from the sids in this file (specify '-' to read file from stdin)

    from_universes : list of str, optional
        create the universe from these existing universes

    exclude_delisted : bool
        exclude delisted securities and expired contracts that would otherwise be
        included (default is not to exclude them)

    append : bool
        append to universe if universe already exists (default False)

    replace : bool
        replace universe if universe already exists (default False)

    Returns
    -------
    dict
        status message

    Examples
    --------
    Create a universe called 'nyse-stk' from a CSV file:

    >>> create_universe("usa-stk", "nyse_securities.csv")

    Create a universe from a DataFrame:

    >>> import io
    >>> f = io.StringIO()
    >>> japan_securities.to_csv(f)
    >>> create_universe("japan-stk", f)
    """
    if append and replace:
        raise ValueError("append and replace are mutually exclusive")

    params = {}
    if from_universes:
        params["from_universes"] = from_universes
    if exclude_delisted:
        params["exclude_delisted"] = exclude_delisted
    if replace:
        params["replace"] = replace

    url = "/master/universes/{0}".format(code)

    if append:
        method = "PATCH"
    else:
        method = "PUT"

    if infilepath_or_buffer == "-":
        response = houston.request(method, url, params=params, data=to_bytes(sys.stdin))

    elif infilepath_or_buffer and hasattr(infilepath_or_buffer, "read"):
        if infilepath_or_buffer.seekable():
            infilepath_or_buffer.seek(0)
        response = houston.request(method, url, params=params, data=to_bytes(infilepath_or_buffer))

    elif infilepath_or_buffer:
        with open(infilepath_or_buffer, "rb") as f:
            response = houston.request(method, url, params=params, data=f)
    else:
        response = houston.request(method, url, params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_create_universe(*args, **kwargs):
    return json_to_cli(create_universe, *args, **kwargs)

def delete_universe(code):
    """
    Delete a universe.

    The listings details of the member securities won't be deleted, only
    their grouping as a universe.

    Parameters
    ----------
    code : str, required
        the universe code

    Returns
    -------
    dict
        status message
    """

    url = "/master/universes/{0}".format(code)

    response = houston.delete(url)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_delete_universe(*args, **kwargs):
    return json_to_cli(delete_universe, *args, **kwargs)

def list_universes():
    """
    List universes and their size.

    Returns
    -------
    dict
        dict of universe:size
    """
    response = houston.get("/master/universes")
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_universes(*args, **kwargs):
    return json_to_cli(list_universes, *args, **kwargs)

def delist_ibkr_security(sid=None, symbol=None, exchange=None, currency=None, sec_type=None):
    """
    Mark an IBKR security as delisted.

    This does not remove any data but simply marks the security as delisted so
    that data services won't attempt to collect data for the security and so
    that the security can be optionally excluded from query results.

    The security can be specified by sid or a combination of other
    parameters (for example, symbol + exchange). As a precaution, the request
    will fail if the parameters match more than one security.

    Parameters
    ----------
    sid : str, optional
        the sid of the security to be delisted

    symbol : str, optional
        the symbol to be delisted (if sid not provided)

    exchange : str, optional
        the exchange of the security to be delisted (if needed to disambiguate)

    currency : str, optional
        the currency of the security to be delisted (if needed to disambiguate)

    sec_type : str, optional
        the security type of the security to be delisted (if needed to disambiguate).
        Possible choices: STK, ETF, FUT, CASH, IND

    Returns
    -------
    dict
        status message
    """
    params = {}
    if sid:
        params["sids"] = sid
    if symbol:
        params["symbols"] = symbol
    if exchange:
        params["exchanges"] = exchange
    if currency:
        params["currencies"] = currency
    if sec_type:
        params["sec_types"] = sec_type

    response = houston.delete("/master/securities/ibkr", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_delist_ibkr_security(*args, **kwargs):
    return json_to_cli(delist_ibkr_security, *args, **kwargs)

def create_ibkr_combo(combo_legs):
    """
    Create an IBKR combo (aka spread), which is a composite instrument consisting
    of two or more individual instruments (legs) that are traded as a single
    instrument.

    Each user-defined combo is stored in the securities master database with a
    SecType of "BAG". The combo legs are stored in the ComboLegs field as a JSON
    array. QuantRocket assigns a sid for the combo consisting of a prefix 'IC'
    followed by an autoincrementing digit, for example: IC1, IC2, IC3, ...

    If the combo already exists, its sid will be returned instead of creating a
    duplicate record.

    Parameters
    ----------
    combo_legs : list, required
        a list of the combo legs, where each leg is a list specifying action, ratio,
        and sid

    Returns
    -------
    dict
        returns a dict containing the generated sid of the combo, and whether a new
        record was created

    Examples
    --------
    To create a calendar spread on VX, first retrieve the sids of the legs:

    >>> from quantrocket.master import download_master_file
    >>> download_master_file("vx.csv", symbols="VIX", exchanges="CFE", sec_types="FUT")
    >>> vx_sids = pd.read_csv("vx.csv", index_col="Symbol").Sid.to_dict()

    Then create the combo:

    >>> create_ibkr_combo([
            ["BUY", 1, vx_sids["VXV9"]],
            ["SELL", 1, vx_sids["VXQ9"]]
        ])
        {"sid": IC1, "created": True}
    """

    f = six.StringIO()
    json.dump(combo_legs, f)
    f.seek(0)

    response = houston.put("/master/combos/ibkr", data=f)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_create_ibkr_combo(combo_filepath):
    with open(combo_filepath) as f:
        combo_legs = json.load(f)
    return json_to_cli(create_ibkr_combo, combo_legs)

def load_rollrules_config(filename):
    """
    Upload a new rollover rules config.

    Parameters
    ----------
    filename : str, required
        the rollover rules YAML config file to upload

    Returns
    -------
    dict
        status message
    """
    with open(filename) as file:
        response = houston.put("/master/config/rollover", data=file.read())
    houston.raise_for_status_with_json(response)
    return response.json()

def get_rollrules_config():
    """
    Returns the current rollover rules config.

    Returns
    -------
    dict
        the config as a dict
    """
    response = houston.get("/master/config/rollover")
    houston.raise_for_status_with_json(response)
    # It's possible to get a 204 empty response
    if not response.content:
        return {}
    return response.json()

def _cli_load_or_show_rollrules(filename=None):
    if filename:
        return json_to_cli(load_rollrules_config, filename)
    else:
        return json_to_cli(get_rollrules_config)

def collect_ibkr_calendar(exchanges=None):
    """
    Collect upcoming trading hours from IBKR for exchanges and save to
    securities master database.

    Parameters
    ----------
    exchanges : list of str, optional
        limit to these exchanges

    Returns
    -------
    dict
        status message
    """
    params = {}
    if exchanges:
        params["exchanges"] = exchanges

    response = houston.post("/master/calendar/ibkr", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_collect_ibkr_calendar(*args, **kwargs):
    return json_to_cli(collect_ibkr_calendar, *args, **kwargs)

def list_calendar_statuses(exchanges, sec_type=None, in_=None, ago=None, outside_rth=False):
    """
    Check whether exchanges are open or closed.

    Parameters
    ----------
    exchanges : list of str, required
        the exchange(s) to check

    sec_type : str, optional
        the security type, if needed to disambiguate for exchanges that
        trade multiple security types. Possible choices: STK, FUT, CASH, OPT

    in_ : str, optional
        check whether exchanges will be open or closed at this point in the
        future (use Pandas timedelta string, e.g. 2h or 30min or 1d)

    ago : str, optional
        check whether exchanges were open or closed this long ago
        (use Pandas timedelta string, e.g. 2h or 30min or 1d)

    outside_rth : bool
        check extended hours calendar (default is to check regular
        trading hours calendar)

    Returns
    -------
    dict
        exchange calendar status
    """
    params = {}
    if exchanges:
        params["exchanges"] = exchanges
    if sec_type:
        params["sec_type"] = sec_type
    if in_:
        params["in"] = in_
    if ago:
        params["ago"] = ago
    if outside_rth:
        params["outside_rth"] = outside_rth

    response = houston.get("/master/calendar", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_calendar_statuses(*args, **kwargs):
    return json_to_cli(list_calendar_statuses, *args, **kwargs)

def _cli_in_status_since(status, since=None, in_=None, ago=None):

    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this command")

    dt = pd.Timestamp.now(status["timezone"])
    if in_:
        dt += pd.Timedelta(in_)
    elif ago:
        dt -= pd.Timedelta(ago)

    dt = dt.tz_localize(None)

    required_since = pd.date_range(periods=5, end=dt,
                                   freq=since, normalize=False)

    # For >1D freq, normalize to midnight
    if required_since.freq.is_anchored() or required_since.freq.rule_code == "D":
        required_since = pd.date_range(periods=5, end=dt, freq=since, normalize=True)
        required_since = required_since[-1]
    else:
        # If not normalized, the last value is dt, so use the penultimate value
        required_since = required_since[-2]

    actual_since = pd.Timestamp(status["since"])
    return actual_since <= required_since

def _cli_in_status_until(status, until=None, in_=None, ago=None):

    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this command")

    dt = pd.Timestamp.now(status["timezone"])
    if in_:
        dt += pd.Timedelta(in_)
    elif ago:
        dt -= pd.Timedelta(ago)

    dt = dt.tz_localize(None)

    required_until = pd.date_range(start=dt, periods=5,
                                   freq=until, normalize=False)

    # For >1D freq, normalize to midnight
    if required_until.freq.is_anchored() or required_until.freq.rule_code == "D":
        required_until = pd.date_range(start=dt, periods=5,
                                   freq=until, normalize=True)
        # due to normalize=True, the date range might include a time before the
        # start dt; filter it out
        required_until = required_until[required_until > dt]
        required_until = required_until[0]
    else:
        # If not normalized, the first value is dt, so use the second value
        required_until = required_until[1]

    actual_until = pd.Timestamp(status["until"])
    return actual_until >= required_until

def _cli_isopen(exchanges, sec_type=None, in_=None, ago=None, since=None, until=None, outside_rth=False):
    statuses = list_calendar_statuses(exchanges, sec_type=sec_type, in_=in_, ago=ago, outside_rth=outside_rth)
    is_open = all([
        calendar["status"] == "open" for calendar in statuses.values()
    ])
    if is_open and since:
        is_open = all([
            _cli_in_status_since(status, since=since, in_=in_, ago=ago)
            for status in statuses.values()])

    elif is_open and until:
        is_open = all([
            _cli_in_status_until(status, until=until, in_=in_, ago=ago)
            for status in statuses.values()])

    return '', int(not is_open)

def _cli_isclosed(exchanges, sec_type=None, in_=None, ago=None, since=None, until=None, outside_rth=False):
    statuses = list_calendar_statuses(exchanges, sec_type=sec_type, in_=in_, ago=ago, outside_rth=outside_rth)
    is_closed = all([
        calendar["status"] == "closed" for calendar in statuses.values()
    ])
    if is_closed and since:
        is_closed = all([
            _cli_in_status_since(status, since=since, in_=in_, ago=ago)
            for status in statuses.values()])

    elif is_closed and until:
        is_closed = all([
            _cli_in_status_until(status, until=until, in_=in_, ago=ago)
            for status in statuses.values()])

    return '', int(not is_closed)

def round_to_tick_sizes(infilepath_or_buffer, round_fields,
                        how=None, append_ticksize=False,
                        outfilepath_or_buffer=None):
    """
    Round prices in a CSV file to valid tick sizes.

    CSV should contain columns `Sid`, `Exchange`, and the columns to be rounded
    (e.g. `LmtPrice`). Additional columns will be ignored and returned unchanged.

    Parameters
    ----------
    infilepath_or_buffer : str or file-like object, required
        CSV file with prices to be rounded (specify '-' to read file from stdin)

    round_fields : list of str, required
        columns to be rounded

    how : str, optional
        which direction to round to. Possible choices: 'up', 'down', 'nearest'
        (default is 'nearest')

    append_ticksize : bool
        append a column of tick sizes for each field to be rounded (default False)

    outfilepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    Returns
    -------
    None
    """

    params = {}
    if round_fields:
        params["round_fields"] = round_fields
    if how:
        params["how"] = how
    if append_ticksize:
        params["append_ticksize"] = append_ticksize

    url = "/master/ticksizes.csv"

    if infilepath_or_buffer == "-":
        # No-op if an empty file is passed on stdin
        f = six.StringIO(sys.stdin.read())
        if not f.getvalue():
            return

        response = houston.get(url, params=params, data=to_bytes(f))

    elif infilepath_or_buffer and hasattr(infilepath_or_buffer, "read"):
        if infilepath_or_buffer.seekable():
            infilepath_or_buffer.seek(0)
        response = houston.get(url, params=params, data=to_bytes(infilepath_or_buffer))

    elif infilepath_or_buffer:
        with open(infilepath_or_buffer, "rb") as f:
            response = houston.get(url, params=params, data=f)
    else:
        raise ValueError("infilepath_or_buffer is required")

    houston.raise_for_status_with_json(response)

    filepath_or_buffer = outfilepath_or_buffer or sys.stdout
    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_round_to_tick_sizes(*args, **kwargs):
    return json_to_cli(round_to_tick_sizes, *args, **kwargs)



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

def add_subparser(subparsers):
    _parser = subparsers.add_parser("ibg", description="QuantRocket IB Gateway service CLI", help="Start and stop IB Gateway")
    _subparsers = _parser.add_subparsers(title="subcommands", dest="subcommand")
    _subparsers.required = True

    examples = """
Set username/password and trading mode (paper/live) for IB Gateway, or view
current username and trading mode.

Can be used to set new credentials or switch between paper and live trading
(must have previously entered live credentials). Setting new credentials will
restart IB Gateway and takes a moment to complete.

Credentials are encrypted at rest and never leave your deployment.

Examples:

View current credentials for IB Gateway service named ibg1 (shows username and
trading mode only):

    quantrocket ibg credentials ibg1

Set credentials for ibg1 (will prompt for password):

    quantrocket ibg credentials ibg1 -u myuser --paper

Leave credentials as-is but switch to live trading (must have previously entered
live credentials):

    quantrocket ibg credentials ibg1 --live
    """
    parser = _subparsers.add_parser(
        "credentials",
        help="set username/password and trading mode (paper/live) for IB Gateway",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "gateway",
        metavar="SERVICE_NAME",
        help="name of IB Gateway service to set credentials for (for example, "
        "'ibg1')")
    parser.add_argument(
        "-u", "--username",
        metavar="USERNAME",
        help="IBKR username (optional if only modifying trading mode)")
    parser.add_argument(
        "-p", "--password",
        metavar="PASSWORD",
        help="IBKR password (if omitted and user is provided, will be prompted "
        "for password)")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--paper",
        action="store_const",
        dest="trading_mode",
        const="paper",
        help="set trading mode to paper trading")
    group.add_argument(
        "--live",
        action="store_const",
        dest="trading_mode",
        const="live",
        help="set trading mode to live trading")
    parser.set_defaults(func="quantrocket.ibg._cli_get_or_set_credentials")

    examples = """
Query statuses of IB Gateways.

Examples:

List the status of all gateways:

    quantrocket ibg status

Get a list of gateways that are running:

    quantrocket ibg status --status running
    """
    parser = _subparsers.add_parser(
        "status",
        help=" query statuses of IB Gateways",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-s", "--status",
        choices=["running", "stopped", "error"],
        help="limit to IB Gateways in this status. Possible choices: %(choices)s")
    parser.add_argument(
        "-g", "--gateways",
        metavar="SERVICE_NAME",
        nargs="*",
        help="limit to these IB Gateways")
    parser.set_defaults(func="quantrocket.ibg._cli_list_gateway_statuses")

    examples = """
Start one or more IB Gateways.

Examples:

Asynchronously start all gateways (that aren't already running):

    quantrocket ibg start

Start specific gateways and wait for them to come up:

    quantrocket ibg start --gateways ibg1 ibg3 --wait

Restart all gateways:

    quantrocket ibg stop --wait && quantrocket ibg start
    """
    parser = _subparsers.add_parser(
        "start",
        help="start one or more IB Gateways",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-g", "--gateways",
        metavar="SERVICE_NAME",
        nargs="*",
        help="limit to these IB Gateways")
    parser.add_argument(
        "-w", "--wait",
        action="store_true",
        default=False,
        help="wait for the IB Gateway  to start before returning (default is to "
        "start the gateways asynchronously)")
    parser.set_defaults(func="quantrocket.ibg._cli_start_gateways")

    examples = """
Stop one or more IB Gateways.

Examples:

Stop all gateways (that aren't already stopped):

    quantrocket ibg stop

Stop specific gateways and wait for them to stop:

    quantrocket ibg stop --gateways ibg1 ibg3 --wait
    """
    parser = _subparsers.add_parser(
        "stop",
        help="stop one or more IB Gateways",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-g", "--gateways",
        metavar="SERVICE_NAME",
        nargs="*",
        help="limit to these IB Gateways")
    parser.add_argument(
        "-w", "--wait",
        action="store_true",
        default=False,
        help="wait for the IB Gateway to stop before returning (default is to stop "
        "the gateways asynchronously)")
    parser.set_defaults(func="quantrocket.ibg._cli_stop_gateways")

    examples = """
Upload a new IB Gateway permissions config, or return the current configuration.

Permission configs are only necessary when running multiple IB Gateways with
differing market data permissions.

Examples:

Upload a new config (replaces current config):

    quantrocket ibg config myconfig.yml

Show current config:

    quantrocket ibg config
    """
    parser = _subparsers.add_parser(
        "config",
        help="upload a new config, or return the current configuration",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "filename",
        nargs="?",
        metavar="FILENAME",
        help="the config file to upload (if omitted, return the current config)")
    parser.set_defaults(func="quantrocket.ibg._cli_load_or_show_config")



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
from quantrocket.houston import houston
from quantrocket.cli.utils.output import json_to_cli

def list_databases(services=None, codes=None, detail=False, expand=False):
    """
    List databases.

    Parameters
    ----------
    services : str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    detail : bool
        return database statistics (default is to return a
        flat list of database names)

    expand : bool
        expand sharded databases to include individual shards
        (default is to list sharded databases as a single database)

    Returns
    -------
    dict
        dict of lists of databases (one key for PostgreSQL databases and one for
        SQLite databases)

    Examples
    --------
    Load database details in a pandas DataFrame:

    >>> from quantrocket.db import list_databases
    >>> import itertools
    >>> databases = list_databases(detail=True)
    >>> databases = pd.DataFrame.from_records(itertools.chain(databases["sqlite"], databases["postgres"]))
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    if detail:
        params["detail"] = detail
    if expand:
        params["expand"] = expand

    response = houston.get("/db/databases", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_databases(*args, **kwargs):
    return json_to_cli(list_databases, *args, **kwargs)

def get_s3_config():
    """
    Return the current S3 configuration, if any.

    See http://qrok.it/h/dbs3 to learn more.

    Returns
    -------
    dict
        configuration details
    """
    response = houston.get("/db/s3config")
    houston.raise_for_status_with_json(response)
    # It's possible to get a 204 empty response
    if not response.content:
        return {}
    return response.json()

def set_s3_config(access_key_id=None, secret_access_key=None, bucket=None):
    """
    Set AWS S3 configuration for pushing and pulling databases to and from
    S3.

    See http://qrok.it/h/dbs3 to learn more.

    Credentials are encrypted at rest and never leave your deployment.

    Parameters
    ----------
    access_key_id : str, optional
        AWS access key ID

    secret_access_key : str, optional
        AWS secret access key (if omitted and access_key_id is provided,
        will be prompted for secret_access_key)

    bucket : str, optional
        the S3 bucket name to push to/pull from

    Returns
    -------
    dict
        status message
    """
    if access_key_id and not secret_access_key:
        secret_access_key = getpass.getpass(prompt="Enter AWS Secret Access Key: ")

    data = {}
    if access_key_id:
        data["access_key_id"] = access_key_id
    if secret_access_key:
        data["secret_access_key"] = secret_access_key
    if bucket:
        data["bucket"] = bucket

    response = houston.put("/db/s3config", data=data)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_or_set_s3_config(access_key_id=None, secret_access_key=None,
                                   bucket=None, *args, **kwargs):
    if access_key_id or secret_access_key or bucket:
        return json_to_cli(set_s3_config, access_key_id, secret_access_key, bucket, *args, **kwargs)
    else:
        return json_to_cli(get_s3_config, *args, **kwargs)

def s3_push_databases(services=None, codes=None):
    """
    Push database(s) to Amazon S3.

    See http://qrok.it/h/dbs3 to learn more.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    Returns
    -------
    json
        status message
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    response = houston.put("/db/s3", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_s3_push_databases(*args, **kwargs):
    return json_to_cli(s3_push_databases, *args, **kwargs)

def s3_pull_databases(services=None, codes=None, force=False):
    """
    Pull database(s) from Amazon S3.

    See http://qrok.it/h/dbs3 to learn more.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    force: bool
        overwrite existing database if one exists (default is to
        fail if one exists)

    Returns
    -------
    json
        status message
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    if force:
        params["force"] = force
    response = houston.get("/db/s3", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_s3_pull_databases(*args, **kwargs):
    return json_to_cli(s3_pull_databases, *args, **kwargs)

def optimize_databases(services=None, codes=None):
    """
    Optimize databases to improve performance.

    This runs the 'VACUUM' command, which defragments the database and
    reclaims disk space.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these service

    codes: list of str, optional
        limit to these codes

    Returns
    -------
    json
        status message
    """
    params = {}
    if codes:
        params["codes"] = codes
    if services:
        params["services"] = services
    response = houston.post("/db/optimizations", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_optimize_databases(*args, **kwargs):
    return json_to_cli(optimize_databases, *args, **kwargs)



# Copyright 2019 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run: python -m unittest discover -s tests/ -p test*.py -t .

import unittest
try:
    from unittest.mock import patch
except ImportError:
    # py27
    from mock import patch
import pandas as pd
import pytz
import numpy as np
from quantrocket.master import get_securities_reindexed_like, get_contract_nums_reindexed_like
from quantrocket.exceptions import ParameterError

class SecuritiesReindexedLikeTestCase(unittest.TestCase):

    @patch("quantrocket.master.download_master_file")
    def test_pass_sids_based_on_reindex_like(self, mock_download_master_file):
        """
        Tests that sids are correctly passed to the download_master_file
        function based on reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                          "FI23456"],
                     Symbol=["ABC",
                             "DEF"],
                     Etf=[1,
                          0],
                     Delisted=[0,
                               1],
                     Currency=["USD",
                               "USD"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        get_securities_reindexed_like(closes, fields=["Symbol", "Etf", "Delisted", "Currency"])

        download_master_file_call = mock_download_master_file.mock_calls[0]
        _, args, kwargs = download_master_file_call
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertListEqual(kwargs["fields"], ["Symbol", "Etf", "Delisted", "Currency"])

    @patch("quantrocket.master.download_master_file")
    def test_cast_boolean_fields(self, mock_download_master_file):
        """
        Tests that master fields Etf and Delisted are cast to boolean.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Symbol=["ABC","DEF"],
                                           Delisted=[0, 1],
                                           Etf=[1, 0],
                                           edi_Delisted=[0,1],
                                           ibkr_Etf=[1, 0]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields=["Symbol", "Etf", "Delisted", "edi_Delisted", "ibkr_Etf"])

        symbols = securities.loc["Symbol"]
        symbols = symbols.reset_index()
        symbols.loc[:, "Date"] = symbols.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            symbols.to_dict(orient="records"),
            [{'Date': '2018-05-01T00:00:00', "FI12345": "ABC", "FI23456": "DEF"},
            {'Date': '2018-05-02T00:00:00', "FI12345": "ABC", "FI23456": "DEF"},
            {'Date': '2018-05-03T00:00:00', "FI12345": "ABC", "FI23456": "DEF"}]
        )

        for field in ("Delisted", "edi_Delisted"):
            delisted = securities.loc[field]
            delisted = delisted.reset_index()
            delisted.loc[:, "Date"] = delisted.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
            self.assertListEqual(
                delisted.to_dict(orient="records"),
                [{'Date': '2018-05-01T00:00:00', "FI12345": False, "FI23456": True},
                {'Date': '2018-05-02T00:00:00', "FI12345": False, "FI23456": True},
                {'Date': '2018-05-03T00:00:00', "FI12345": False, "FI23456": True}]
            )

        for field in ("Etf", "ibkr_Etf"):
            etfs = securities.loc[field]
            etfs = etfs.reset_index()
            etfs.loc[:, "Date"] = etfs.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
            self.assertListEqual(
                etfs.to_dict(orient="records"),
                [{'Date': '2018-05-01T00:00:00', "FI12345": True, "FI23456": False},
                {'Date': '2018-05-02T00:00:00', "FI12345": True, "FI23456": False},
                {'Date': '2018-05-03T00:00:00', "FI12345": True, "FI23456": False},
                ]
            )

    def test_securities_reindexed_like(self):
        """
        Tests get_securities_reindexed_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01",
                                periods=3,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     Symbol=["ABC",
                             "DEF"],
                     Etf=[1,
                          0],
                     Delisted=[0,
                               1],
                     Currency=["USD",
                               "EUR"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields=["Symbol", "Etf", "Delisted", "Currency"])

            securities = securities.reset_index()
            securities.loc[:, "Date"] = securities.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
            self.assertListEqual(
                securities.to_dict(orient="records"),
                [{'Field': 'Currency',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Currency',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Currency',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Delisted',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Delisted',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Delisted',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Etf',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Etf',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Etf',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Symbol',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'},
                 {'Field': 'Symbol',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'},
                 {'Field': 'Symbol',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'}]
            )

    def test_securities_reindexed_like_intraday(self):
        """
        Tests get_securities_reindexed_like with Date and Time in the index.
        """
        closes = pd.DataFrame(
            np.random.rand(4,2),
            columns=["FI12345","FI23456"],
            index=pd.MultiIndex.from_product([
                pd.date_range(start="2018-05-01",
                              periods=2,
                              freq="D",
                              tz="America/New_York",
                              name="Date"),
                ["09:30:00", "09:31:00"],
                ], names=["Date", "Time"]))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     Symbol=["ABC",
                             "DEF"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields="Symbol")

            securities = securities.reset_index()
            securities.loc[:, "Date"] = securities.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")

            self.assertListEqual(
                securities.to_dict(orient="records"),
                [{"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-01T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:30:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-01T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:31:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-02T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:30:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-02T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:31:00'}]
            )
class ContractNumsReindexedLikeTestCase(unittest.TestCase):

    @patch("quantrocket.master.download_master_file")
    def test_pass_sids_based_on_reindex_like(self, mock_download_master_file):
        """
        Tests that sids are correctly passed to the download_master_file
        function based on reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     ibkr_UnderConId=[1,
                                 1],
                     SecType=["FUT",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-07-04"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        get_contract_nums_reindexed_like(closes)

        download_master_file_call = mock_download_master_file.mock_calls[0]
        _, args, kwargs = download_master_file_call
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertListEqual(kwargs["fields"], ["RolloverDate", "ibkr_UnderConId", "SecType"])

    def test_complain_if_no_fut(self):
        """
        Tests error handling when you pass a DataFrame with no FUTs.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01",
                                periods=3,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                    dict(Sid=["FI12345",
                                "FI23456"],
                         ibkr_UnderConId=[1,
                                     1],
                         SecType=["STK",
                                  "STK"],
                         RolloverDate=[
                             None,
                             None
                         ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            with self.assertRaises(ParameterError) as cm:
                get_contract_nums_reindexed_like(closes)

        self.assertIn("input DataFrame does not appear to contain any futures contracts",
                      repr(cm.exception))

    def test_contract_nums_reindexed_like(self):
        """
        Tests get_contract_nums_reindexed_like.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 3.0},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_contract_nums_reindexed_like_tz_naive(self):
        """
        Tests get_contract_nums_reindexed_like with a tz-naive reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 3.0},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_limit_sequence_num(self):
        """
        Tests get_contract_nums_reindexed_like with the limit option.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes, limit=2)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 'nan'},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 'nan'
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 'nan'},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_contract_nums_reindexed_like_intraday(self):
        """
        Tests get_contract_nums_reindexed_like when the input DataFrame includes
        Dates and Times.
        """
        closes = pd.DataFrame(
            np.random.rand(8,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.MultiIndex.from_product([
                pd.date_range(start="2018-05-01",
                              periods=4,
                              freq="D",
                              tz="America/New_York",
                              name="Date"),
                ["09:30:00","09:31:00"],
            ], names=["Date","Time"]))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        self.assertEqual(list(contract_nums.index.names), ["Date","Time"])

        contract_nums = contract_nums.reset_index().fillna("nan")
        contract_nums.loc[:, "Date"] = contract_nums.Date.dt.strftime("%Y-%m-%d")

        self.assertListEqual(
            contract_nums.to_dict(orient="records"),
            [{"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-01',
              'Time': '09:30:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-01',
              'Time': '09:31:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-02',
              'Time': '09:30:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-02',
              'Time': '09:31:00'},
             {"FI12345": 'nan',
              "FI23456": 2.0,
              "FI34567": 1.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-03',
              'Time': '09:30:00'},
             {"FI12345": 'nan',
              "FI23456": 2.0,
              "FI34567": 1.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-03',
              'Time': '09:31:00'},
             {"FI12345": 'nan',
              "FI23456": 1.0,
              "FI34567": 1.0,
              "FI45678": 'nan',
              "FI56789": 'nan',
              "FI67890": 2.0,
              'Date': '2018-05-04',
              'Time': '09:30:00'},
             {"FI12345": 'nan',
              "FI23456": 1.0,
              "FI34567": 1.0,
              "FI45678": 'nan',
              "FI56789": 'nan',
              "FI67890": 2.0,
              'Date': '2018-05-04',
              'Time': '09:31:00'}]
        )
